/*
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright (c) 2011 digitalSTROM.org, Zurich, Switzerland
 *  Author: Roman Köhler <roman.koehler@aizo.com>
 *  Complete rewrite by: Andreas Brauchli <andreas.brauchli@aizo.com>
 */

var LOGFILE_NAME = "scene-responder.log";
var LOG = new Logger(LOGFILE_NAME);

function pad(c) {
    var i = parseInt(c, 10);
    return (i < 10 ? '0' + i : c.toString());
}

/**
 * Check if a scene is supported by this scene responder version
 * @param s {Number} Scene id to check
 * @return {Boolean} True if supported, false if unsupported
 */
function validScene(s) {
    if ((s >= 10 && s <= 16) || (s >= 32 && s <= 63) || s >= 74) {
        return false;
    }
    return true;
}

/**
 * Returns an unused unique identifier (id)
 * @return {String} A unique string
 */
function getId() {
    var ndTriggers = Property.getNode('triggers'),
        triggers,
        max = 0;

    if (ndTriggers) {
        triggers = ndTriggers.getChildren();
        while (ndTriggers.getChild(max.toString())) {
            ++max;
        }
        return max.toString();

    } else {
        return "0";
    }
}

function startup() {
    // Prepare app
    Property.load();
    LOG.logln("scene-responder running");

    // - THIS IS ONLY A MIGRATION TEST CASE -
    // Example property of old version 0.1.4
    // Property.getNode('triggers').getParent().removeChild('triggers');
    // Property.setProperty('triggers/treppenhaus/params', "{\"name\":\"treppenhaus\",\"trigger\":\"{\\\"scene\\\":5,\\\"zone\\\":1314,\\\"group\\\":1}\",\"actions\":\"{\\\"actions\\\":[{\\\"unit\\\":\\\"scene\\\",\\\"group\\\":1,\\\"zone\\\":1314,\\\"scene\\\":0}]}\",\"duration\":\"1\"}");
    // Property.setProperty('triggers/test/params', '{\"name\":\"test\",\"trigger\":\"{\\\"scene\\\":14,\\\"zone\\\":0,\\\"group\\\":1}\",\"actions\":\"{\\\"actions\\\":[{\\\"unit\\\":\\\"group\\\",\\\"group\\\":1,\\\"zone\\\":1314,\\\"value\\\":61}]}\",\"duration\":\"5\"}');
    // convertOldTriggers();
    // - END OF MIGRATION TEST CASE -

    var ndVersion = Property.getNode('version');
    if (ndVersion) {
        var version = ndVersion.getValue().split('.');
        if (version.length > 1 && version[0] == '0' && parseInt(version[1], 10) < 4) {
            convertOldTriggers();
        }
    }

    if (!Property.getNode('triggers')) {
        createNode('triggers');
    }
    //set version number
    Property.setProperty('version', '0.6.0.2');
    Property.setFlag('version', 'ARCHIVE', true);
    Property.store();
}

/**
 * A function that converts stored scene responder triggers from the old to the new version
 */
function convertOldTriggers() {
    LOG.logln('Converting triggers from old scene responder app');
    Property.load();
    var ndTriggers = Property.getNode('triggers');
    if (!ndTriggers) {
        return;
    }
    var triggers = ndTriggers.getChildren();
    var i;
    for (i = 0; i < triggers.length; ++i) {
        var ndOldParams = triggers[i].getChild('params'),
            oldParams,
            oldTrigger,
            oldActions;

        if (!ndOldParams) {
            LOG.logln('Error: no old params found to convert on trigger ' + triggers[i].name);
            deleteOldTrigger(triggers[i]);
            continue;
        }

        try {
            // LOG.logln('Debug: old params: ' + ndOldParams.getValue());
            oldParams = JSON.parse(ndOldParams.getValue());
            // LOG.logln('Debug: old trigger: ' + decodeURIComponent(oldParams.trigger));
            oldTrigger = JSON.parse(oldParams.trigger);
            // LOG.logln('Debug: old actions: ' + decodeURIComponent(oldParams.actions));
            oldActions = JSON.parse(oldParams.actions);
        } catch (e) {
            LOG.logln('Deleting old trigger '+ triggers[i].name +': failed to parse');
            deleteOldTrigger(triggers[i]);
            continue;
        }

        //
        // validity checks
        //

        if (!oldTrigger
            || isNaN(oldTrigger.scene)
            || isNaN(oldTrigger.zone)
                // allowed to be unset in very old versions (< 0.1.4)
            || (oldTrigger.group && isNaN(oldTrigger.group))
        ) {
            // delete invalid triggers
            LOG.logln('Deleting old trigger '+ triggers[i].name +': missing trigger');
            deleteOldTrigger(triggers[i]);
            continue;
        }

        if (!validScene(oldTrigger.scene)) {
            LOG.logln('Deleting old trigger '+ triggers[i].name +': scene ' + oldTrigger.scene +' is not supported by this version');
            deleteOldTrigger(triggers[i]);
            continue;
        }

        if (!oldActions
            || !oldActions.actions
            || !(oldActions.actions.length > 0)
        ) {
            // delete invalid triggers
            LOG.logln('Deleting old trigger '+ triggers[i].name +': missing actions');
            deleteOldTrigger(triggers[i]);
            continue;
        }
        oldActions = oldActions.actions;

        //
        // Conversions
        //

        // new params object
        var params = {
            name: triggers[i].name, // or oldParams.name
            disabled: false, // there existed no possibility to disable an action
            scene: parseInt(oldTrigger.scene, 10),
            zone: parseInt(oldTrigger.zone, 10)
        };

        // Convert delay
        if (oldParams.duration === undefined) {
            params.delay = 0;
        } else if (isNaN(oldParams.duration)) {
            LOG.logln('Warning: invalid delay on trigger '+ triggers[i].name +', assuming 0');
            params.delay = 0;
        } else {
            params.delay = 60 * parseInt(oldParams.duration, 10);
        }

        // Convert trigger group
        if (isNaN(oldTrigger.group)) {
            LOG.logln('Warning: no trigger group is set on '+ triggers[i].name +': defaulting to group 1 (Lights)');
            params.group = 1;
        } else {
            params.group = parseInt(oldTrigger.group, 10);
        }

        // Convert old actions
        params.actions = [];
        var a;
        for (a = 0; a < oldActions.length; ++a) {
            var p = {};
            if (isNaN(oldActions[a].zone)
                || (isNaN(oldActions[a].scene) && isNaN(oldActions[a].value))
                || (oldActions[a].group && isNaN(oldActions[a].group))
                || (!isNaN(oldActions[a].value)
                    && (oldActions[a].value < 0 || oldActions[a].value > 255))
            ) {
                LOG.logln('Error: no or invalid action set on trigger '+ triggers[i].name +' action at index '+ a +'. Skipping');
                continue;
            }
            p.zone = parseInt(oldActions[a].zone, 10);
            if (oldActions[a].group === undefined) {
                LOG.logln('Warning: no group is set on '+ triggers[i].name +' action at index '+ a +', defaulting to group 1 (Lights)');
                p.group = 1;
            } else {
                p.group = parseInt(oldActions[a].group, 10);
            }
            if (oldActions[a].scene !== undefined) {
                if (!validScene(oldActions[a].scene)) {
                    LOG.logln('Warning unsupported scene '
                            + oldActions[a].scene +' call from action '+ a
                            +' on '+ triggers[i].name +': skipping');
                    continue;
                }

                p.scene = parseInt(oldActions[a].scene, 10);
            } else if (oldActions[a].value !== undefined) {
                p.outputValue = parseInt(oldActions[a].value, 10);
            }
            params.actions.push(p);
        }
        if (params.actions.length === 0) {
            LOG.logln('Error: no valid actions on trigger '+ triggers[i].name +': deleting');
            deleteOldTrigger(triggers[i]);
            continue;
        }

        // store converted trigger
        deleteOldTrigger(triggers[i]);
        addTrigger(params);
        LOG.logln('Converted trigger '+ params.idname);
    }
    Property.store();
} // convertOldTriggers

/**
 * Execute actions delayed one by one
 * @param actions {Array} Actions to execute (in the order of the array)
 * @param time {Number} Miliseconds to wait between actions
 * @param callback {Function} (optional) callback to execute after the last action execution
 */
function executeActions(actions, time, callback) {
    var action = actions.shift(),
        // filter target by presence to avoid dSS crash (bug #1261)
        target = getDevices().byPresence(true),
        targetlog = '',
        callNextAction = function () { executeActions(actions, time, callback); };

    if (!isNaN(action.group)) {
        targetlog += ' on zone ' + action.zone + ', group ' + action.group;
        target = target.byZone(action.zone).byGroup(action.group);

    } else if (action.device) {
        targetlog += ' on device ' + action.device;
        target = target.byDSID(action.device);
    }

    if (target && target.length() > 0) {
        if (action.scene !== undefined) {
            LOG.logln('Exec: calling scene ' + action.scene + targetlog);
            target.callScene(parseInt(action.scene, 10));
        }
        else if (action.blink) {
            LOG.logln('Exec: blinking' + targetlog);
            target.blink();
        }
        else if (action.outputValue !== undefined) {
            LOG.logln('Exec: setting ouput ' + action.outputValue + targetlog);
            target.setValue(parseInt(action.outputValue, 10));
        }
    } else {
        LOG.logln('Warning: empty target (devices probably offline)');
    }

    if (actions.length > 0) {
        // LOG.logln('Debug: Scheduling next '+ actions.length +' params in '+ time +'ms');
        setTimeout(callNextAction, time);
    } else if (callback !== undefined) {
        callback();
    }
} // executeActions

/**
 * Fetches the parameters of this trigger from the property tree.
 * @param ndTrigger {Object} Property tree node of the trigger
 * @return {Object} Object containing the read out values of the trigger
 */
function loadParams(ndTrigger) {
    // LOG.logln('Debug: loading params for node with id: ' + ndTrigger.name);
    var name = ndTrigger.getChild('name');
    if (!name) {
        LOG.logln('Error: no trigger name');
        return null;
    }

    var params = {
        id: ndTrigger.name,
        idname: ndTrigger.name + ' (' + name.getValue() + ')',
        name: name.getValue(),
        delay: parseInt(ndTrigger.getChild('delay').getValue(), 10),
        scene: parseInt(ndTrigger.getChild('scene').getValue(), 10),
        zone: parseInt(ndTrigger.getChild('zone').getValue(), 10),
        group: parseInt(ndTrigger.getChild('group').getValue(), 10),
        actions: []
    };
    var ndDisabled = ndTrigger.getChild('disabled');
    params.disabled = (ndDisabled ? !! ndDisabled.getValue() : false);

    var ndQueueId = ndTrigger.getChild('queueId');
    if (ndQueueId) {
        params.queueId = ndQueueId.getValue();
    }

    // LOG.logln('Debug: Loading actions');
    var ndActions = ndTrigger.getChild('actions');
    var loadValue = function(action, ndAction, name) {
        var c = ndAction.getChild(name);
        if (c) {
            action[name] = c.getValue();
            return true;
        }
        return false;
    };
    if (ndActions) {
        var actions = ndActions.getChildren();
        var a;
        for (a = 0; a < actions.length; ++a) {
            var curAction = {};
            if (loadValue(curAction, actions[a], 'device')) {}
            else if (loadValue(curAction, actions[a], 'group') && loadValue(curAction, actions[a], 'zone')) {}
            else {
                LOG.logln('Error: no target on an action '+ params.idname);
                continue;
            }
            if (loadValue(curAction, actions[a], 'blink')) {}
            else if (loadValue(curAction, actions[a], 'scene')) {}
            else if (loadValue(curAction, actions[a], 'outputValue')) {}
            else {
                LOG.logln('Error: invalid scene on action '+ params.idname);
                continue;
            }
            params.actions.push(curAction);
            // LOG.logln('Debug: loaded action: '+ JSON.stringify(curAction));
        }
    } else {
        LOG.logln('Error: trigger '+ params.idname +' has no actions set');
    }

    // LOG.logln('Debug: params loaded');
    return params;
}

/**
 * Updates a node if it exists or creates a new archived property node otherwise
 * @param path {String} Relative path
 * @param value {String|Boolean|Integer} Value to set
 */
function updateOrCreateNode(path, value) {
    var nd = Property.getNode(path);
    if (nd) {
        nd.setValue(value);
    } else {
        createNode(path, value);
    }
}

/**
 * Creates a new archived property node with given value
 * @param path {String} Relative path
 * @param value {String|Boolean|Integer} Value to set
 * @return {Object} The newly created node
 */
function createNode(path, value) {
    var nd = new Property(path);
    if (value !== undefined) {
        nd.setValue(value);
    }
    nd.setFlag('ARCHIVE', true);
    return nd;
}

/**
 * Stores _VALID_ parameters into the property tree.
 * Use checkParams(params) to validate them first!
 * Make sure to call Property.store() after calling this method
 * @param params {Object} Object to store
 * @return {Object} The trigger node
 */
function storeParams(params) {
    var triggerPath = 'triggers/' + params.id;
    var actionPath = triggerPath + '/actions';
    var i, c;

    var ndTrigger = Property.getNode(triggerPath) || createNode(triggerPath);

    updateOrCreateNode(triggerPath + '/id', params.id);
    updateOrCreateNode(triggerPath + '/name', params.name);
    updateOrCreateNode(triggerPath + '/delay', params.delay);
    updateOrCreateNode(triggerPath + '/disabled', !!params.disabled);
    updateOrCreateNode(triggerPath + '/scene', params.scene);
    updateOrCreateNode(triggerPath + '/group', params.group);
    updateOrCreateNode(triggerPath + '/zone', params.zone);

    // actions
    var ndActions = ndTrigger.getChild('actions');
    if (!ndActions) {
        ndActions = createNode(actionPath);
    } else {
        // clean out actions, as indices might not match up
        for (c in ndActions.getChildren()) {
            ndActions.removeChild(c);
        }
    }

    // populate actions
    for (i = 0; i < params.actions.length; ++i) {
        var p = actionPath + '/' + i;
        createNode(p);

        if (params.actions[i].zone !== undefined) {
            createNode(p + '/zone', params.actions[i].zone);
        }
        if (params.actions[i].group !== undefined) {
            createNode(p + '/group', params.actions[i].group);
        }
        if (params.actions[i].device) {
            createNode(p + '/device', params.actions[i].device);
        }
        if (params.actions[i].blink) {
            createNode(p + '/blink', true);
        }
        if (params.actions[i].scene !== undefined) {
            createNode(p + '/scene', params.actions[i].scene);
        }
        if (params.actions[i].outputValue !== undefined) {
            createNode(p + '/outputValue', params.actions[i].outputValue);
        }
    }
    return ndTrigger;
} // storeParams

/**
 * Checks if parameters conform to what is expected
 * @param triggerParams {Object} Object to check
 * @return {Object} Conform object or null on error
 */
function checkParams(triggerParams) {
    if (!triggerParams.name) {
        LOG.logln('Error: trigger has no name');
        return null;
    }

    if (!triggerParams.actions || triggerParams.actions.length === 0) {
        LOG.logln('Error: trigger has no actions');
        return null;
    }

    var actions = [];
    var i;
    for (i = 0; i < triggerParams.actions.length; ++i) {
        var a = triggerParams.actions[i];
        if (!a.device && isNaN(a.group)) { // isNan(undefined) === true
            LOG.logln('Error: no or invalid group or device defined as target on action ' + i);
            return null;
        }
        if (a.blink === undefined && a.scene === undefined && a.outputValue === undefined) {
            LOG.logln('Error: no action defined at index ' + i);
            return null;
        }
        if (
            (a.scene !== undefined && isNaN(a.scene))
            || (a.group !== undefined && isNaN(a.group))
            || (a.outputValue !== undefined && (
                isNaN(a.outputValue) || a.outputValue < 0 || a.outputValue > 255
            ))
        ) {
            LOG.logln('Error: invalid action defined at index ' + i);
            return null;
        }

        var action = {};
        if (a.zone !== undefined) {
            action.zone = parseInt(a.zone, 10);
        }
        if (a.group !== undefined) {
            action.group = parseInt(a.group, 10);
        }
        if (a.device !== undefined) {
            action.device = a.device;
        }
        if (a.blink) {
            action.blink = true;
        }
        if (a.scene !== undefined) {
            action.scene = parseInt(a.scene, 10);
        }
        if (a.outputValue !== undefined) {
            action.outputValue = parseInt(a.outputValue, 10);
        }
        actions.push(action);
    }

    if (isNaN(triggerParams.delay)) {
        LOG.logln('Warning: invalid delay on action ' + i +' assuming 0');
        triggerParams.delay = 0;
    }

    if (isNaN(triggerParams.scene)
        || isNaN(triggerParams.group)
        || isNaN(triggerParams.zone)
    ) {
        LOG.logln('Error: invalid trigger on action ' + i);
        return null;
    }

    // create the valid params object
    var params = {
        name: triggerParams.name,
        delay: parseInt(triggerParams.delay, 10),
        scene: parseInt(triggerParams.scene, 10),
        group: parseInt(triggerParams.group, 10),
        zone: parseInt(triggerParams.zone, 10),
        disabled: !!triggerParams.disabled,
        actions: actions
    };
    // add id, idname fields if they were supplied
    if (triggerParams.id) {
        params.id = triggerParams.id;
        params.idname = params.id + ' (' + params.name + ')';
    }

    return params;
} // checkParams

/**
 * Add a new trigger
 * @param params {Object} The new trigger is created from these parameters
 */
function addTrigger(triggerParams) {
    // check triggerParams and set params for serialization
    var params = checkParams(triggerParams),
        // tmpId is used to communicate the id of the newly created trigger
        // to the client. Since there is no way to specify a return value
        // for an event, the client will be able to specify a temporary id
        // tmpId and poll for that id. When it's created, it'll contain the
        // definitive id of the created trigger.
        // The client is charged with deleting this node
        tmpId;

    if (params) {
        if (params.id === undefined || params.id.substr(0, 3) === 'tmp') {
            if (params.id !== undefined) {
                tmpId = params.id;
            }
            params.id = getId();
            params.idname = params.id + ' (' + params.name + ')';
        }

        storeParams(params);

        if (tmpId !== undefined) {
            var evt = new Event('scene-responder.create.' + tmpId, { id: params.id });
            evt.raise();
        }

        Property.store();
    }
}

function executeTrigger(ndTrigger, params) {
    if (!params.disabled) {
        LOG.logln(
            'Processing trigger ' + params.idname
            + ' (' + params.actions.length +' actions)'
        );
        executeActions(params.actions, 500, function() {
            // remove queueId after execution sequence
            if (params.queueId) {
                ndTrigger.removeChild('queueId');
                delete params.queueId;
            }
        });
    }
}

function dequeueTrigger(ndTrigger, params) {
    if (params.queueId) {
        LOG.logln('Debug: dequeueing trigger '+ params.idname);
        var qNode = Property.getNode('/system/EventInterpreter/ScheduledEvents');
        qNode.removeChild(params.queueId);

        ndTrigger.removeChild('queueId');
        delete params.queueId;
    }
}

function deleteOldTrigger(ndTrigger) {
    ndTrigger.getParent().removeChild(ndTrigger);
    Property.store();
}

function deleteTrigger(ndTrigger, params) {
    LOG.logln('Delete: trigger '+ params.idname);

    dequeueTrigger(ndTrigger, params);
    ndTrigger.getParent().removeChild(ndTrigger);
    Property.store();
}

/**
 * Enable or disable a scene response trigger
 * @param enable {Boolean} True to enable, false to disable
 */
function enableTrigger(ndTrigger, params, enable) {
    LOG.logln((enable ? 'Enabling' : 'Disabling') + ' trigger ' + params.idname);

    params.disabled = !enable;
    updateOrCreateNode('triggers/' + params.id + '/disabled', !enable);
    if (!enable) {
        dequeueTrigger(ndTrigger, params);
    }
    Property.store();
}

function sceneCalled() {
    // these need to be parsed as they are strings!
    var sceneId = parseInt(raisedEvent.parameter.sceneID, 10),
        zoneId = parseInt(raisedEvent.source.zoneID, 10),
        groupId = parseInt(raisedEvent.source.groupID, 10);

    // LOG.logln('Debug: scene '+ sceneId + ' called from zone: ' + zoneId
    //            + (groupId ? (', group: ' + groupId) : ''));

    Property.load();

    var ndTriggers = Property.getNode('triggers');
    if (!ndTriggers) {
        return;
    }
    var triggers = ndTriggers.getChildren();
    var i;
    for (i = 0; i < triggers.length; ++i) {
        var params = loadParams(triggers[i]);
        if (!params || params.disabled) {
            continue;
        }
        // LOG.logln('Debug:         scene: '+ params.scene
        //            + ' zone: ' + params.zone + ' group: ' + params.group);

        if (params.scene === sceneId
            && (( // The zone is set to broadcast
                 zoneId === 0
                       // .. then accept these 4 scenes for any set zone or group
                 && (sceneId === 65     /* Panic */
                     || sceneId === 71  /* Present */
                     || sceneId === 72  /* Absent */
                     || sceneId === 73) /* Bell */
                )
                  // Or the zone matches
                || ((params.zone === zoneId || params.zone === 0)
                     && ( // .. then accept apartment scenes from all groups
                         sceneId >= 64
                          // .. and the group also matches
                         || (params.group === groupId || params.group === 0))))
        ) {
            dequeueTrigger(triggers[i], params);
            if (params.delay > 0) {
                var evt = new TimedEvent(
                    'scene-responder',
                    '+' + params.delay,
                    {
                        action_type: 'execute',
                        trigger_id: params.id
                    }
                );
                params.queueId = evt.raise();
                // update prop tree (don't store as it's not restart-persistent)
                var qId = triggers[i].getChild('queueId');
                if (qId) {
                    qId.setValue(params.queueId);
                } else {
                    Property.setProperty('triggers/' + params.id + '/queueId', params.queueId);
                }
                // LOG.logln('Debug: action delayed for ' + params.delay + ' seconds');
            } else {
                executeTrigger(triggers[i], params);
            }
        }
    }
}

function main() {
    if (raisedEvent.name == 'running') {
        startup();
        return;

    } else if (raisedEvent.name == 'callScene') {
        sceneCalled();
        return;
    }

    /*
    LOG.logln('Debug: scene-responder main call with action type: '
                + raisedEvent.parameter.action_type
                + (raisedEvent.parameter.params === undefined ? ''
                    : ' and params: '
                    + raisedEvent.parameter.params.toString()
                )
    );
    */
    var params;
    if (raisedEvent.parameter.action_type == 'add') {
        // Add a new trigger
        if (raisedEvent.parameter.params === undefined) {
            LOG.logln('Error: missing trigger parameters');
            return;
        }
        LOG.logln('Adding trigger with parmeters: '+ raisedEvent.parameter.params);
        try {
            params = JSON.parse(raisedEvent.parameter.params);
            addTrigger(params);
        } catch(e) {
            LOG.logln('Error: Failed to parse params');
        }

    } else {
        // Operations that involve a specific trigger's parameters
        Property.load();
        var ndTrigger = Property.getNode('triggers/' + raisedEvent.parameter.trigger_id);
        if (ndTrigger) {
            params = loadParams(ndTrigger);

            if (raisedEvent.parameter.action_type == 'enable') {
                enableTrigger(ndTrigger, params, true);
            }
            else if (raisedEvent.parameter.action_type == 'disable') {
                enableTrigger(ndTrigger, params, false);
            }
            else if (raisedEvent.parameter.action_type == 'delete') {
                deleteTrigger(ndTrigger, params);
            }
            else if (raisedEvent.parameter.action_type == 'execute') {
                executeTrigger(ndTrigger, params);
            }
        } else {
            LOG.logln('Error: node with id '+ raisedEvent.parameter.trigger_id +' doesn\'t exist');
        }
    }
} // main

main();
